/*
TODO:
	- 01. Property Window: Keep scroll position for each layout type
	- 02. ? Property Window: Collapsable sections (Should work with above feature)
	- 03. Per project/gui grid colour
	- 04. Recent Files list
	- ? Grid in Tab2
	- Properties to complete (Editor and export):
		- !!!! Check that the Text property works correctly for all types of controls eg. ListView should create columns/headers
		- !!!! Check Styles/ExStyles work correctly for all types of controls
		- Uncommon: Styles/ExStyles (Also see SmartGUI and especially Gui-Creator)
		- Text
		- Edit
		- UpDown
		- Picture
		- Button
		- Checkbox
		- Radio
		- DropDownList
			- Redraw issues with tabs
		- ComboBox
		- ListBox
		- ListView
			- ListView +LV0x00 extended styles
		- TreeView
		- Link
		- Hotkey
		- DateTime
		- MonthCal
		- Slider
		- Progress
		- GroupBox
		- Tab2
			- Uses window BG colour instead of window control colour (Also depends on "Background" property)
			- ? Tab icons, see bookmarked page (http://www.autohotkey.com/board/topic/5692-icons-in-tabs/)
			- How to handle editing tabs (Adding/Removing)?
			- Export
				- Add "Change Tab Control Order" Option (Default=Yes)
		- StatusBar
		- ActiveX (e.g. Internet Explorer Control)
		- Custom
	- ? Add other controls such as RichEdit
	- 
	- ? UpDown: Increments other than 1
	- ? UpDown: Syncing UpDown and buddy value
	- Test in 32bit AHK (Check all DllCall/Struct/SendMessage)
	- DEBUG: Reset FileLoad to show the file select dialog
	- ListView and TreeViewitem editor
		- Track updating certain properties (Sort, Multi, etc.)
	- ? Anchor library
	- ? Trim certain proeprty values when updated
	- ? Add more icons to the status bar to display different settings, using different part numbers (If there is any more useful information that can be dispalyed there)
	- ? Undo, Redo system
		- Current file modified state - Save, Load, New, Exit
	- ? Gui control explorer or some other way of finding controls that are off screen
		- Scrollbars ?
	- ? Menu designer
	- ? BUG: Freeze when opening Tab2 Text sub win
*/

; -- Setup script directives
; -----------------------------------------------------------------
{

#Include includes\Gdip_All.ahk
#Include includes\Gdip_Ext.ahk
#Include includes\GuiButtonIcon.ahk
#Include includes\HD_EX.ahk
#Include includes\ChooseColor.ahk
#Include includes\Dlg_Font.ahk
; Debug
#Include includes\Debug.ahk
#Warn, , StdOut

#NoEnv
#Persistent
;~ #Warn UseUnsetLocal, Off
#Warn LocalSameAsGlobal, Off
CoordMode Menu, Screen
CoordMode Mouse, Screen
CoordMode ToolTip, Screen
DetectHiddenWindows On
SendMode Input
SetBatchLines -1
SetControlDelay -1
SetWinDelay -1
SetWorkingDir %A_ScriptDir%
StringCaseSense On

}

; -- Initialize variables
; -----------------------------------------------------------------
{

global APP_NAME					:= "Simple Gui Editor"
	 , CONTROL_MIN_SIZE			:= 0
	 , RESIZE_HANDLE_SIZE		:= 6
	 , RESIZE_HANDLE_HALF_SIZE	:= RESIZE_HANDLE_SIZE / 2
	 , SETTINGS_FILE			:= "settings.ini"
	 
	 , GUI_DELIMITER	:= Chr(0x007)
	 , MAX_INT			:=  2147483647
	 , MIN_INT			:= -2147483648

global GdipToken
	 , Settings
	 
	 ; CurrentParent: The parent control that is currently being edited (for the moment nothing or a Tab2 control).
	 ; Any control that is created will belong to this parent control. 0 = No parent/Gui window
	 ; To support tab pages this value is a combination of two values: The parent control index (bytes 1 and 2), and the page index (bytes 3 and 4)
	 , CurrentParent := 0, CurrentParentHwnd := 0
	 , ParentControls := {0: {Visible: 1}}
	 
	 , CurrentGui := 0			; Stores the data for the current gui, options, controls, etc.
	 , CurrentControlData := 0
	 , UniqueControls := {}
	 , CustomMouseDownEvents := {}
	 , MainMenuAccKeys := {}	; Maps menu paths (eg. File>Save) to the actual menu text which is needed by AHK to identify the menu item
	 , ErrorMessageFunc := Func("ErrorMessage")
	 , CurrentAppAction := 0
	 , CursorOverlayWindow
	 
	 , ExportOptionGuiName, ExportOptionGuiNameLabel, ExportOptionGuiLabel, ExportOptionEOL
	 , ExportOptionComma, ExportOptionNamedConstants := 0, ExportOptionNamedConstantsList := {}
	
	; >> Settings window <<
	
	, SettingsWinHwnd

#Include Win32Constants.ahk
#Include Properties.ahk
#Include PropertyClass.ahk
#Include CursorOverlayWindowClass.ahk

}

; -- Auto execute
; -----------------------------------------------------------------
{

OnExit OnAppExit

#Include Utilities.ahk
#Include MainWin.ahk
#Include EditorWin.ahk
#Include PropertyWin.ahk

if(!GdipToken := Gdip_Startup()){
	MsgBox, 48, gdiplus error!, Gdiplus failed to start. Please ensure you have gdiplus on your system
	ExitApp
}

SettingsLoad()
InitializePropertyDefinitions()
ResetCurrentGuiData()
CreateGui()

;~ FileLoad()

} return

; -- Functions
; -----------------------------------------------------------------

AddControl(ControlHwnd, Type, ControlData := 0){
	ControlId := CurrentGui.NextIndex
	while(StrLen(ControlId) < 3)
		ControlId := "0" ControlId
	ControlId := Type ControlId
	
	PropertyClass.ControlType := Type
	
	if(!ControlData){
		GuiControlGet Control, Editor:Pos, %ControlHwnd%
		CurrentGui.Controls[ControlHwnd] := ControlData := {ControlType: Type, Text: ControlTypeOptions[Type].Text, Parent: 0}
		
		Property := PropertiesDef.Common.Position
		if(!Property.Get("Disable"))
			ControlData.Position := [ControlX, ControlY]
		
		Property := PropertiesDef.Common.Size
		if(!Property.Get("Disable"))
			ControlData.Size := [Property.Get("StartEmptyX") ? "" : ControlW, Property.Get("StartEmptyY") ? "" : ControlH]
	}
	else
		CurrentGui.Controls[ControlHwnd] := ControlData
	
	if(CurrentParent){
		Property := PropertiesDef.Common.Parent
		if(!Property.Get("Disable")){
			ParentControlAddChild(CurrentParent, CurrentGui.NextIndex)
			ControlData.Parent := CurrentParent
		}
	}
	
	ControlData.Id := ControlId
	ControlData.Index := CurrentGui.NextIndex
	
	CurrentGui.ControlIndexes[CurrentGui.NextIndex++] := ControlHwnd
	CurrentGui.ControlCount++
	
	return ControlData
}

AlignControls(AlignType){
	Min := AlignType = "LeftEdges" or AlignType = "TopEdges" ? 999999 : 0
	
	ControlCount := 0
	for ControlHwnd in Selection.Controls{
		ControlData := CurrentGui.Controls[ControlHwnd]
		ControlX := ControlData.Position[1]
		ControlY := ControlData.Position[2]
		ControlW := ControlData.Size[1]
		ControlH := ControlData.Size[2]
		
		if(AlignType = "LeftEdges")
			Min := ControlX < Min ? ControlX : Min
		else if(AlignType = "HorizontalCenters")
			Min += ControlX + ControlW / 2
		else if(AlignType = "RightEdges")
			Min := ControlX + ControlW > Min ? ControlX + ControlW : Min
		else if(AlignType = "TopEdges")
			Min := ControlY < Min ? ControlY : Min
		else if(AlignType = "VerticalCenters")
			Min += ControlY + ControlH / 2
		else if(AlignType = "BottomEdges")
			Min := ControlY + ControlH > Min ? ControlY + ControlH : Min
		
		ControlCount++
	}
	
	if(AlignType = "HorizontalCenters" or AlignType = "VerticalCenters")
		Min /= ControlCount
	
	for ControlHwnd in Selection.Controls{
		ControlData := CurrentGui.Controls[ControlHwnd]
		ControlX := ControlData.Position[1]
		ControlY := ControlData.Position[2]
		ControlW := ControlData.Size[1]
		ControlH := ControlData.Size[2]
		
		if(AlignType = "LeftEdges")
			ControlX := Min
		else if(AlignType = "HorizontalCenters")
			ControlX := Round(Min - ControlW / 2)
		else if(AlignType = "RightEdges")
			ControlX := Min - ControlW
		else if(AlignType = "TopEdges")
			ControlY := Min
		else if(AlignType = "VerticalCenters")
			ControlY :=  Round(Min - ControlH / 2)
		else if(AlignType = "BottomEdges")
			ControlY := Min - ControlH
		
		GuiControl Editor:Move, %ControlHwnd%, x%ControlX% y%ControlY%
		
		ControlData.Position[1] := ControlX
		ControlData.Position[2] := ControlY
		EditorWinUpdateHighlight(ControlHwnd)
	}
	
	EditorWinUpdateResizeHandles()
	OnPropertyWinPropertyChange("Common", "Position")
}

BuildControlOption(ControlData, ByRef Options, PropertyGroup, PropertyName, Property := "", PropertyOutputName := ""){
	static PropertyValues := [], PropertyValuesIndex := 0, PropertyDefaults := []
	
	if(!IsObject(Property))
		PropertyType := Property, Property := PropertiesDef[PropertyGroup][PropertyName], PropertyClass.ControlType := PropertyType
	if(PropertyOutputName = "")
		PropertyOutputName := PropertyName
	
	if((OnlyIfCondition := Property.Get("OnlyIf"))
	and GetPropertyValue(PropertyGroup, OnlyIfCondition[1], ControlData) != OnlyIfCondition[2])
		return
	
	Type := Property.Get("Type")
	Default := Property.Get("Default")
	
	if(Type = "Text" or Type = "TextToggle"){
		Value :=  GetPropertyValue(PropertyGroup, PropertyName, ControlData)
		Text := Type = "TextToggle" ? Value[2] : Value
		if((Split := Property.Get("OutputSplit"))){
			NewText := ""
			Loop Parse, Text, %Split%
				if(A_LoopField != "")
					NewText .= " " PropertyOutputName A_LoopField
			Text := Trim(NewText), PropertyOutputName := ""
		}	
		
		if(Value != "" and (Type != "TextToggle" or Value[1] = "Yes"))
			Options .= " " PropertyOutputName Text
	}
	
	else if(Type = "Toggle"){
		if((Value := GetPropertyValue(PropertyGroup, PropertyName, ControlData)) != Default)
			Options .= " " (Default = "Yes" or Default = "" and Value = "No" ? "-" : "") Property.Get("OutputPrefix")
			. ((ConstantName := Property.Get("Constant")) ? GetConstant(ConstantName) : PropertyOutputName)
	}
	
	else if(Type = "Int"){
		if((Value :=  GetPropertyValue(PropertyGroup, PropertyName, ControlData)) != "" and !Property.Get("OutputIgnore" Value))
			Options .= " " PropertyOutputName Value
	}
	
	else if(Type = "IntToggle"){
		Value :=  GetPropertyValue(PropertyGroup, PropertyName, ControlData)
		Ticked := Value[1], Value := Value [2]
		if(Ticked != Default)
			Options .= " " (Default = "Yes" or Default = "" and Ticked = "No" ? "-" : "") PropertyOutputName (Ticked != "" ? Value : "")
	}
	
	else if(Type = "Label" or Type = "MultiText" or Type = "Var"){
		Value := GetPropertyValue(PropertyGroup, PropertyName, ControlData)
		if(Value != "")
			Options .= " " PropertyOutputName Value
	}
	
	else if(Type = "List"){
		if((Value := GetPropertyValue(PropertyGroup, PropertyName, ControlData)) != Default){
			Index := Property.Get("ListIndexes")[Value]
			OutputValue := Index and (OptionValues := Property.Get("OptionValues")) ? OptionValues[Index] : Value
			if(Property["Constant" Index])
				OutputValue := GetConstant(OutputValue)
			
			if((OutputValue := Property.Get("OutputPrefix") OutputValue) != "")
				Options .= " " OutputValue
		}
	}

	else if(Type = "XY"){
		Value := GetPropertyValue(PropertyGroup, PropertyName, ControlData)
		if(Property.Get("OutputAsDimensions")){
			if(Value[1] != "" or Value[2] != "")
				Options .= " " PropertyOutputName Value[1] "x" Value[2]
		}
		else if((OutputPrefix := Property.Get("OutputPrefix"))){
			if(Value[1] != "")
				Options .= " " OutputPrefix[1] Value[1]
			if(Value[2] != "")
				Options .= " " OutputPrefix[2] Value[2]
		}
		else if(Value[1] != "" or Value[2] != "")
			Options .= " " PropertyOutputName (Value[1] != "" ? Value[1] : MIN_INT) "-" (Value[2] != "" ? Value[2] : MAX_INT)
	}
	
	else if(Type = "Date"){
		Value :=  GetPropertyValue(PropertyGroup, PropertyName, ControlData)
		if(Property.Get("NoTime") and Value != "")
			FormatTime Value, %Value%, yyyyMMdd
		if(Value != "")
			Options .= " " PropertyOutputName Value
	}
	
	else if(Type = "DateRange"){
		Date :=  GetPropertyValue(PropertyGroup, PropertyName, ControlData, "Raw")
		DateMin := SubStr(Date[1], 2)
		DateMax := SubStr(Date[2], 2)
		
		if(Property.Get("NoTime")){
			if(DateMin != "")
				FormatTime DateMin, %DateMin%, yyyyMMdd
			if(DateMax != "")
				FormatTime DateMax, %DateMax%, yyyyMMdd
		}
		
		IsDateRange := DateMax != "" and DateMin != ""
		NotEmpty := DateMax != "" or DateMin != ""
		Value := DateMin (IsDateRange or NotEmpty and !Property.Get("AllowSingle") ? "-" : "") DateMax
		if(Value != "")
			Options .= " " PropertyOutputName Value
	}
	
	else if(Type = "Composite"){
		CompositeProperties := Property.Get("Properties")
		CompositeFunction := Property.Get("CompositeFunction")
		
		PropertyValuesIndex := 0
		for Index, CompositePropertyName in CompositeProperties{
			CompositeProperty := PropertiesDef[PropertyGroup][CompositePropertyName]
			CompositeValue := GetPropertyValue(PropertyGroup, CompositePropertyName, ControlData)
			PropertyValues[++PropertyValuesIndex] := (OutputValue := CompositeProperty.Get("OutputValue")) ? OutputValue : CompositeValue
			PropertyDefaults[PropertyValuesIndex] := (CompositeValue = CompositeProperty.Get("Default"))
		}
		
		if(CompositeFunction = "Add"){
			Value := 0, Index := 1
			while(Index <= PropertyValuesIndex){
				if(!PropertyDefaults[Index])
					Value += PropertyValues[Index]
				Index++
			}
		}
		
		else if(CompositeFunction = "Range"){
			if(PropertyValues[1] != PropertyDefaults[1] or PropertyValues[2] != PropertyDefaults[2])
				Value := PropertyValues[1] (PropertyValues[2] != "" ? Property.Get("JoinChar") : "") PropertyValues[2]
			else
				Value := ""
		}
		
		if(Value)
			Options .= " " PropertyOutputName Value
	}
}

BuildControlOptions(ControlData, ByRef Options, UseDefaultColour := 1, FullOutput := 1){
	ControlType := ControlData.ControlType
	if(Options != "")
		Options .= " "
	
	C := ControlData.C = "" ? (UseDefaultColour and !ControlTypeOptions[ControlType].NoGuiTextColour ? GetPropertyValue("Gui", "C", CurrentGui.Window) : "") : ControlData.C
	if(C != "")
		Options .= " c" C
	
	PropertyGroup := PropertiesDef.Common
	
	for Index, PropertyName in PropertyGroup.GroupIndexes{
		Property := PropertyGroup[PropertyName], PropertyClass.ControlType := ControlType
		OutputName := Property.Get("OutputName")
		if(Property.Get("CustomOutput") and !OutputName
		or Property.Get("FullOutputOnly") and !FullOutput
		or FullOutput and Property.Get("EditorOnly")
		or Property.Get("Disable"))
			continue
		
		BuildControlOption(ControlData, Options, "Common", PropertyName, Property, OutputName)
	}
	
	PropertyGroup := PropertiesDef[ControlType]
	for Index, PropertyName in PropertyGroup.GroupIndexes{
		Property := PropertyGroup[PropertyName], PropertyClass.ControlType := ControlType
		OutputName := Property.Get("OutputName")
		if(Property.Get("CustomOutput") and !OutputName
		or Property.Get("FullOutputOnly") and !FullOutput
		or FullOutput and Property.Get("EditorOnly")
		or Property.Get("Disable"))
			continue
		
		BuildControlOption(ControlData, Options, ControlType, PropertyName, Property, OutputName)
	}
	
	if(ControlType = "Picture"){
		if(ControlData.IconIndex != "" and ControlData.IconIndex > 0)
			Options .= " Icon" ControlData.IconIndex
	}
}

BuildFontOptions(Font, ByRef FontOptions){
	FontOptions := ""
	
	if(Font.Size)
		FontOptions .= "s" Font.Size " "
	FontStyleOptions := ""
	if(InStr(Font.Style, "bold"))
		FontStyleOptions .= "bold "
	if(InStr(Font.Style, "italic"))
		FontStyleOptions .= "italic "
	if(InStr(Font.Style, "underline"))
		FontStyleOptions .= "underline "
	if(InStr(Font.Style, "strikeout"))
		FontStyleOptions .= "strike "
	if(FontStyleOptions != "")
		FontStyleOptions := "norm " FontStyleOptions
	
	FontOptions .= FontStyleOptions
}

BuildGuiFontOptions(ControlData, ByRef Output := 0, PropertyGroup := "Common", IncludeColour := ""){
	static CurrentFontName, CurrentFontSize, CurrentFontStyle
		 , GuiFont, NextFont := {}
	
	if((Export := Output != 0))
		Output := ""
	
	if(ControlData = "Reset"){
		CurrentFontName := CurrentFontSize := CurrentFontStyle := ""
		GuiFont := GetPropertyValue("Gui", "Font", CurrentGui.Window)
		if(!Export)
			Gui %ExportOptionGuiNameLabel%Font
		return
	}
	
	Font := GetPropertyValue(PropertyGroup, "Font", ControlData)
	NextFont.Name := Font.Name, NextFont.Size := Font.Size, NextFont.Style := Font.Style
	NextFont.Name  := NextFont.Name  = "" ? GuiFont.Name  : NextFont.Name
	NextFont.Size  := NextFont.Size  = "" ? GuiFont.Size  : NextFont.Size
	NextFont.Style := NextFont.Style = "" ? GuiFont.Style : NextFont.Style
	if(NextFont.Name != CurrentFontName or NextFont.Size != CurrentFontSize or NextFont.Style != CurrentFontStyle or IncludeColour){
		CurrentFontName := NextFont.Name, CurrentFontSize := NextFont.Size, CurrentFontStyle := NextFont.Style
		if(CurrentFontName CurrentFontStyle CurrentFontStyle IncludeColour != ""){
			BuildFontOptions(NextFont, FontOptions)
			if(IncludeColour)
				FontOptions .= "c" IncludeColour
			if(!Export){
				if(NextFont.Name)
					Gui %ExportOptionGuiNameLabel%Font, %FontOptions%, % NextFont.Name
				else
					Gui %ExportOptionGuiNameLabel%Font, %FontOptions%
			}
			else if(NextFont.Name)
				Output .= "Gui " ExportOptionGuiNameLabel "Font, " FontOptions ", " NextFont.Name ExportOptionEOL
			else
				Output .= "Gui " ExportOptionGuiNameLabel "Font, " FontOptions ExportOptionEOL
		}
		else if(!Export)
			Gui %ExportOptionGuiNameLabel%Font
		else
			Output .= "Gui " ExportOptionGuiNameLabel "Font" ExportOptionEOL
	}
}

CleanMenuItemName(ItemName){
	/*
	By default menu items with shortcut keys will include those shortcut keys and other extra characters in the name making them hard to identify.
	This function will strip out the shortcut name, space, tabs and & characters to produce a string that is easily identifiable, eg:
		A menu item name that might look like this: 'Delete this &Item`tCtrl+Delete'
		would return 'DeleteThisItem'
	*/
	
	StringUpper ItemName, ItemName, T
	ItemName := RegExReplace(RegExReplace(ItemName, "\&([^&])", "$1"), " +", "")
	KeyIndex := InStr(ItemName, "`t", false, 0)
	if(KeyIndex)
		ItemName := SubStr(ItemName, 1, KeyIndex - 1)
	
	return ItemName
}

CloseApplication(){
	; Still deciding on what this function is going to do
	
	;~ MsgBox, 262179, SimpleGui Editor, Save changes to ??
	;~ IfMsgBox Yes
		;~ out("Yes")
	;~ IfMsgBox No
		;~ out("No")
	;~ IfMsgBox Cancel
		;~ return
	
	SettingsSave()
	Gdip_Shutdown(GdipToken)
	ExitApp
}

CreateGui(){
	SysGet SM_CXVSCROLL, 2
	SysGet SM_CXHSCROLL, 21
	
	Menu, Tray, Icon, images\application.ico, , 1
	
	CursorOverlayWindow := new CursorOverlayWindowClass()
	MainWinCreateGui()
	EditorWinCreateGui()
	PropertyWinCreateGui()
	SettingsWinCreateGui()
	
	GroupAdd AppWindows, ahk_id %MainWinHwnd%
	GroupAdd AppWindows, ahk_id %EditorWinHwnd%
	GroupAdd AppWindows, ahk_id %PropertyWinHwnd%
	GroupAdd EditorWinGroup, ahk_id %EditorWinHwnd%
	
	OnMessage(WM_LBUTTONDOWN, "OnMouseButtonDown")
	Hotkey IfWinActive, ahk_group AppWindows
	Hotkey ~*WheelUp, OnMouseWheel
	Hotkey ~*WheelDown, OnMouseWheel
	Hotkey IfWinActive
}

CreateNewGui(){
	if(CurrentGui.Modified){
		MsgBox 3, % GetPropertyValue("Gui", "Title", CurrentGui.Window), Would you like to save your changes before closing?
		IfMsgBox Yes
			FileSave()
		IfMsgBox Cancel
			return
	}
	
	EditorWinSetCurrentParent(0)
	ParentControls := {0: {Visible: 1}}
	EditorWinSelectControl(0, "clear")
	
	for ControlHwnd, ControlData in CurrentGui.Controls
		DeleteControl(ControlHwnd, 0, 0)
	
	ResetCurrentGuiData()
	EditorWinResetWindow()
	PropertyWinUpdateGroups(1)
	
	for ControlType, Data in CurrentGui.UnqiueInstance
		if(!UniqueControls[Data.Hwnd])
			Data.Hwnd := 0
}

DeleteControl(ControlHwnd, Delelect := 1, ClearData := 1){
	ControlData := CurrentGui.Controls[ControlHwnd]
	
	if(ClearData){
		if(ControlData.UpDownBuddy)
			CurrentGui.Controls[ CurrentGui.ControlIndexes[ControlData.UpDownBuddy] ].Target := ""
		
		ControlType := ControlData.ControlType
		if(ControlType = "UpDown"){
			if(ControlData.Target)
				CurrentGui.Controls[ CurrentGui.ControlIndexes[ControlData.Target] ].UpDownBuddy := ""
		}
		else if(ControlType = "Radio"){
			ControlData.Group := ""
			EditorWinUpdateRadioGroup(ControlData)
		}
		
		if(Delelect and Selection.Controls[ControlHwnd])
			EditorWinSelectControl(ControlHwnd, "remove")
		
		CurrentGui.ControlIndexes.Remove(ControlData.Index, "")
		CurrentGui.Controls.Remove(ControlHwnd, "")
		CurrentGui.ControlCount--
		
		for PageNumber, Value in ControlData._HasChildrenOnPage
			ParentControlClearChildren(ControlData.Index | (PageNumber << 16))
		if(ControlData.Parent)
			ParentControlAddChild(ControlData.Parent, ControlData.Index, 0)
		if(CurrentParentHwnd = ControlHwnd)
			EditorWinSetCurrentParent(0)
	}
	
	if(!ControlTypeOptions[ControlData.ControlType].IsUnique)
		DllCall("DestroyWindow", Ptr,ControlHwnd)
	else{
		CurrentGui.UnqiueInstance[ControlData.ControlType].Visible := 0
		GuiControl Editor:Hide, %ControlHwnd%
		GuiControl Editor:Disable, %ControlHwnd%
	}
}

DistributeControls(DistributeType){
	SortedControlsHwnd := [], SortedControlsX := []
	Min := 999999, Max := -999999
	
	ControlCount := 0
	for ControlHwnd in Selection.Controls{
		ControlData := CurrentGui.Controls[ControlHwnd]
		ControlX := ControlData.Position[1]
		ControlY := ControlData.Position[2]
		ControlW := ControlData.Size[1]
		ControlH := ControlData.Size[2]
		
		if(DistributeType = "LeftEdges")
			X := ControlX
		else if(DistributeType = "HorizontalCenters")
			X := ControlX + ControlW / 2
		else if(DistributeType = "RightEdges")
			X := ControlX + ControlW
		else if(DistributeType = "TopEdges")
			X := ControlY
		else if(DistributeType = "VerticalCenters")
			X := ControlY + ControlH/ 2
		else if(DistributeType = "BottomEdges")
			X := ControlY + ControlH
		
		if(X < Min)
			Min := X
		if(X > Max)
			Max := X
		
		Index := BinarySearchLessThan(SortedControlsX, X), Index := Index ? Index : 1
		SortedControlsHwnd.Insert(Index, ControlHwnd)
		SortedControlsX.Insert(Index, X)
		
		ControlCount++
	}
	
	Spacing := (Max - Min) / (ControlCount - 1)
	
	Position := Min
	for X, ControlHwnd in SortedControlsHwnd{
		ControlData := CurrentGui.Controls[ControlHwnd]
		ControlX := ControlData.Position[1]
		ControlY := ControlData.Position[2]
		ControlW := ControlData.Size[1]
		ControlH := ControlData.Size[2]
		
		if(DistributeType = "LeftEdges")
			ControlX := Round(Position)
		else if(DistributeType = "HorizontalCenters")
			ControlX := Round(Position - ControlW / 2)
		else if(DistributeType = "RightEdges")
			ControlX := Round(Position - ControlW)
		else if(DistributeType = "TopEdges")
			ControlY := Round(Position)
		else if(DistributeType = "VerticalCenters")
			ControlY := Round(Position - ControlY / 2)
		else if(DistributeType = "BottomEdges")
			ControlY := Round(Position - ControlY)
		
		GuiControl Editor:Move, %ControlHwnd%, x%ControlX% y%ControlY%
		
		ControlData.Position[1] := ControlX
		ControlData.Position[2] := ControlY
		EditorWinUpdateHighlight(ControlHwnd)
		
		Position += Spacing
	}
	
	EditorWinUpdateResizeHandles()
	OnPropertyWinPropertyChange("Common", "Position")
}

ErrorMessage(Message){
	Gui Main:+OwnDialogs
	SetModalMode()
	MsgBox 8208, %APP_NAME%, %Message%
	SetModalMode(0)
	WinActivate ahk_id %MainWinHwnd%
}

FileExport(Preview := 0){
	; > Initialiaze/Gui New
	{
		ExportOptionGuiLabel		:= GetPropertyValue("GuiOptions", "Label", CurrentGui.Window)
		ExportOptionGuiName			:= GetPropertyValue("GuiOutputOptions", "GuiName", CurrentGui.Window)
		ExportOptionGuiNameLabel	:= ""
		ExportOptionEOL				:= GetPropertyValue("GuiOutputOptions", "EOL", CurrentGui.Window)
		ExportOptionEOL				:= ExportOptionEOL = "Windows (\r\n)" ? "`r`n" : ExportOptionEOL = "Unix (\n)" ? "`n" : "`r"
		ExportOptionComma			:= GetPropertyValue("GuiOutputOptions", "CommandComma", CurrentGui.Window) = "Yes" ? ", " : " "
		ExportOptionNamedConstants	:= GetPropertyValue("GuiOutputOptions", "NamedConstants", CurrentGui.Window) = "Yes"
		GuiNameLabel				:= ExportOptionGuiName != "" ? ExportOptionGuiName ":" : ""
		CommentSections				:= GetPropertyValue("GuiOutputOptions", "CommentSections", CurrentGui.Window) = "Yes"
		ExtraComments				:= GetPropertyValue("GuiOutputOptions", "ExtraComments", CurrentGui.Window) = "Yes"
		GuiEventLabel				:= ExportOptionGuiLabel != ""
			? ExportOptionGuiLabel : ExportOptionGuiName != ""
				? ExportOptionGuiName "Gui"
				: "Gui"
		GuiDelimiter				:= CurrentGui.Delimiter, GuiDelimiter := GuiDelimiter = "`t" ? "``t" : GuiDelimiter = "`n" ? "``n" : GuiDelimiter
		ChangeUpDownOrder			:= GetPropertyValue("GuiOutputOptions", "ChangeUpDownOrder", CurrentGui.Window) = "Yes"
		
		Output := ""
		if(CommentSections)
			Output .= "; -- Auto-Execute" ExportOptionEOL ExportOptionEOL
	}
	
	; > Gui New options
	{
		GuiNewOptions := ""
		PropertyGroup := PropertiesDef.GuiOptions
		for Index, PropertyName in PropertyGroup.GroupIndexes{
			Property := PropertyGroup[PropertyName], PropertyClass.ControlType := "Gui"
			if(Property.Get("CustomOutput") or Property.Get("Disable") or Property.Get("EditorOnly"))
				continue
			
			BuildControlOption(CurrentGui.Window, GuiNewOptions, "GuiOptions", PropertyName, Property)
		}
		
		Output .= "Gui" ExportOptionComma GuiNameLabel "New" (GuiNewOptions != "" ? "," GuiNewOptions : "") ExportOptionEOL
		
		BuildGuiFontOptions("Reset", 1)
		FontOptions := ""
		BuildGuiFontOptions(CurrentGui.Window, FontOptions, "Gui", CurrentGui.Window.C)
		Output .= FontOptions
		
		if(CurrentGui.Window.BackgroundColour != "" or CurrentGui.Window.ControlColour){
			BgColour := CurrentGui.Window.BackgroundColour, CtrlColour := CurrentGui.Window.ControlColour
			Output .= "Gui" ExportOptionComma "Color, " BgColour (CtrlColour != "" ? ", " CtrlColour : "") ExportOptionEOL
		}
	}
	
	; > Controls
	
	PreviousControl := 0
		, CurrentRadioGroup := "", FirstGroup := 1
		, OutputHwnd := ""
		, ControlIndexes := [] , Index := 1
		, ControlsOutputData := {}
		, SubroutinesList := {}
		, ControlOutputStack := [], ControlOutputStackIndex := 0
		, TextProperty := PropertiesDef.Common.Text
	
	for ControlId, ControlHwnd in CurrentGui.ControlIndexes
		ControlIndexes[Index++] := CurrentGui.Controls[ControlHwnd].Index
	
	Index := 1, NextControlHwnd := 0
	for ControlId, ControlHwnd in CurrentGui.ControlIndexes{
		ControlOutputStack[++ControlOutputStackIndex] := ControlHwnd
		
		while(ControlOutputStackIndex > 0){
			ControlHwnd := ControlOutputStack[ControlOutputStackIndex--]
			ControlData := CurrentGui.Controls[ControlHwnd]
			ControlType := ControlData.ControlType
			CommandComments := ""
			
			; Wait to output the UpDown after its target/buddy
			if(ChangeUpDownOrder and ControlType = "UpDown" and ControlData.Target and ControlData.Target > ControlData.Index and PreviousControl.Index != ControlData.Target)
				continue
			
			OutputData := ControlsOutputData[ControlData.Index]
			if(!OutputData)
				OutputData := ControlsOutputData[ControlData.Index] := {}
			if(OutputData.Complete)
				continue
			
			; Build generic control options
			{
				Options := ""
				BuildGuiFontOptions(ControlData, FontOptions)
				BuildControlOptions(ControlData, Options, 0)
			}
			
			; Other options
			{
				if(ControlData.UpDownBuddy and ControlData.UpDownBuddy != ControlIndexes[Index + 1]){
					if(ChangeUpDownOrder)
						ControlOutputStack[++ControlOutputStackIndex] := CurrentGui.ControlIndexes[ControlData.UpDownBuddy]
					else
						OutputHwnd := ControlData.Id
				}
				
				if(ControlType = "UpDown"){
					UpDownTargetIndex := ControlData.Target & 0xFFFF
					if(!UpDownTargetIndex or UpDownTargetIndex != PreviousControl.Index){
						if(UpDownTargetIndex)
							OutputHwnd := ControlData.Id
						
						Options .= " -" GetConstant("UDS_AUTOBUDDY")
					}
					else if(UpDownTargetIndex and UpDownTargetIndex = PreviousControl.Index)
						Options .= " " GetConstant("UDS_AUTOBUDDY")
					if(UpDownTargetIndex)
						Options .= " -" GetConstant("UDS_ALIGNRIGHT")
				}
				
				else if(ControlType = "Radio"){
					if(ControlData.Group != CurrentRadioGroup or CurrentRadioGroup = ""){
						if(ExtraComments and ControlData.Group != "")
							CommandComments .= " New radio group [" ControlData.Group "]"
						if(PreviousControl.ControlType = "Radio")
							Options .= " Group"
						CurrentRadioGroup := ControlData.Group
						
						if((RadioGroup := CurrentGui.RadioGroups[ControlData.Group])){
							; Add them to the stack in reverse order so that they are output in the same order that they were created
							Index := RadioGroup.Count
							for ControlIndex in RadioGroup.Controls
								ControlOutputStack[ControlOutputStackIndex + Index--] := CurrentGui.ControlIndexes[ControlIndex]
							ControlOutputStackIndex += RadioGroup.Count
						}
					}
				}
				
				if(ControlData.Hwnd)
					OutputHwnd := ControlData.Hwnd
			}
			
			; Output the control
			{
				if(OutputHwnd){
					Options .= " Hwnd" OutputHwnd
					OutputData.Hwnd := OutputHwnd
					OutputHwnd := ""
				}
				
				PropertyClass.ControlType := ControlType
				if(!Property.Get("OutputNoText")){
					ControlText := GetPropertyValue("Common", "Text", ControlData)
					
					if(TextProperty.Get("Type") = "List"){
						if((OptionValues := TextProperty.Get("OptionValues")))
							Index := TextProperty.Get("ListIndexes")[ControlText]
							, ControlText := Index ? OptionValues[Index] : ControlText
					}
					
					StringReplace ControlText, ControlText, ``, ````, 1
					StringReplace ControlText, ControlText, `n, ``n, 1
					if(TextProperty.Get("GuiDelimiter"))
						StringReplace ControlText, ControlText, %GUI_DELIMITER%, %GuiDelimiter%, 1
				}
				else
					ControlText := ""
				
				Output .= FontOptions
				Output .= "Gui" ExportOptionComma "Add, " ControlType ","
					. (Options != "" ? Options : "")
					. (ControlText != "" ? ", " ControlText : "")
					. (CommandComments != "" ? " `;" CommandComments : "") ExportOptionEOL
			}
			
			; Build post control options
			{
				if((Name := GetPropertyValue("Common", "G", ControlData, 2)) != "")
					SubroutinesList[Name] := ""
				
				if(ControlData.UpDownBuddy and ControlData.Index > ControlData.UpDownBuddy and (UpDownHwnd := ControlsOutputData[ControlData.UpDownBuddy].Hwnd)){
					Output .= "`tSendMessage" ExportOptionComma GetConstant("UDM_SETBUDDY") ", %" ControlData.Id "%, 0, , ahk_id %" UpDownHwnd "%" ExportOptionEOL
				}
				
				if(ControlType = "UpDown"){
					; UpDownTargetIndex is set above in the "Other options" section
					if(UpDownTargetIndex and UpDownTargetIndex != PreviousControl.Index and UpDownTargetIndex < PreviousControl.Index){
						TargetHwnd := ControlsOutputData[UpDownTargetIndex].Hwnd
						OutputData.Hwnd := ControlData.Id
						Output .= "`tSendMessage" ExportOptionComma GetConstant("UDM_SETBUDDY") ", %" TargetHwnd "%, 0, , ahk_id %" ControlData.Id "%" ExportOptionEOL
					}
				}
				
				else if(ControlType = "Tab2"){
					Output .= "Gui" ExportOptionComma "Tab" ExportOptionEOL
				}
			}
			
			PreviousControl := ControlData
			OutputData.Complete := 1
		}
		
		Index++
	}
	
	; > Gui Show
	{
		GuiShowOptions := ""
		for Index, GroupName in ["Gui", "GuiShowOptions"]{
			PropertyGroup := PropertiesDef[GroupName]
			for Index, PropertyName in PropertyGroup.GroupIndexes{
				Property := PropertyGroup[PropertyName], PropertyClass.ControlType := "Gui"
				if(Property.Get("CustomOutput") or Property.Get("Disable") or Property.Get("EditorOnly"))
					continue
				
				BuildControlOption(CurrentGui.Window, GuiShowOptions, GroupName, PropertyName, Property)
			}
		}
		
		Output .= "Gui" ExportOptionComma "Show" (GuiShowOptions != "" ? "," GuiShowOptions : "")
		; Title
		Title := GetPropertyValue("Gui", "Title", CurrentGui.Window)
		Output .= (Title != "" ? ", " Title : "") ExportOptionEOL
	}
	
	; End of auto-exec section
	Output .= "return" ExportOptionEOL ExportOptionEOL
	out(Trim(Output)), Preview := 1
	
	; > Subroutines
	{
		if(Preview or GetPropertyValue("GuiOutputOptions", "GuiCloseEvent", CurrentGui.Window) = "Yes")
			SubroutinesList[GuiEventLabel "Close"] := "ExitApp"
		if(Preview or GetPropertyValue("GuiOutputOptions", "GuiEscapeEvent", CurrentGui.Window) = "Yes")
			SubroutinesList[GuiEventLabel "Escape"] := "ExitApp"
		
		OutputSubroutines := ""
		for Label, Commands in SubroutinesList{
			OutputSubroutines .= Label ":" ExportOptionEOL
			OutputSubroutines .= "`t" Commands ExportOptionEOL
			OutputSubroutines .= "return" ExportOptionEOL ExportOptionEOL
		}
		
		if(OutputSubroutines != ""){
			if(CommentSections)
				Output .= "; -- Events" ExportOptionEOL ExportOptionEOL
			Output .= OutputSubroutines
		}
	}
	
	; > Win32 Constants declarations
	{
		ConstantsOutput := ""
		for ConstantName in ExportOptionNamedConstantsList
			ConstantsOutput .= (ConstantsOutput != "" ? "`t , " : "") ConstantName " := " Win32Constants[ConstantName] ExportOptionEOL
		
		if(ConstantsOutput != ""){
			ConstantsOutput := "global " ConstantsOutput
			if(CommentSections)
				ConstantsOutput := "; -- Win32 Constants" ExportOptionEOL ExportOptionEOL ConstantsOutput
			Output := ConstantsOutput "`n" Output
		}
	}
	
	;~ out(Output), Preview := 1
	if(Preview)
		Output := "#SingleInstance Force" ExportOptionEOL ExportOptionEOL Output
	
	if(Preview){
		ExportFile := A_Temp "\GuiEditor.ahk"
		FileDelete %ExportFile%
		FileAppend %Output%, %ExportFile%
		Run %ExportFile%
	}
	else{
		SetModalMode()
		Gui Main:+OwnDialogs
		FileName := GetPropertyValue("Gui", "Title", CurrentGui.Window) ".ahk"
		FileSelectFile FilePath, S18, %FileName%, Export Gui, Ahk Script (*.ahk)
		SetModalMode(0)
		
		if(!ErrorLevel){
			if(SubStr(FilePath, -3) != ".ahk")
				FilePath .= ".ahk"
			FileDelete %FilePath%
			FileAppend %Output%, %FilePath%
		}
	}
	
	; Reset
	
	ExportData := 0
	ExportOptionNamedConstants := 0
	ExportOptionNamedConstantsList := {}
}

FileLoad(FilePath := ""){
	;~ if(CurrentGui.Modified){
		;~ MsgBox 3, %APP_NAME%, Would you like to save your changes before closing?
		;~ IfMsgBox Yes
			;~ Save()
		;~ IfMsgBox Cancel
			;~ return
	;~ }
	
	;~ SetModalMode()
	;~ Gui Main:+OwnDialogs
	;~ FileSelectFile FilePath, 3, , Load File, Gui File (*.gui)
	;~ SetModalMode(0)
	
	if(ErrorLevel)
		return
		
	FilePath := "C:\My Documents\Untitled GUI.gui"
	
	File := FileOpen(FilePath, "r", "UTF-8")
	if(!IsObject(File)){
		ErrorMessage("There was an error opening the file")
		return
	}
	
	; Clear current Gui and load the data from the file
	EditorWinSetCurrentParent(0)
	ParentControls := {0: {Visible: 1}}
	EditorWinSelectControl(0, "clear")
	
	for ControlHwnd, ControlData in CurrentGui.Controls
		DeleteControl(ControlHwnd, 0, 0)
	
	UnqiueInstance := CurrentGui.UnqiueInstance
	CurrentGui := FileReadObject(File)
	CurrentGui.UnqiueInstance := UnqiueInstance
	Window := CurrentGui.Window
	File.Close()
	
	; Recreate the loaded gui
	
	Gui Editor:Default
	CurrentGui.Path := FilePath
	EditorWinResetWindow(CurrentGui.Window.Size[1], CurrentGui.Window.Size[2])
	ExportOptionGuiNameLabel := "Editor:"
	EditorWinCreateControlFromData("Reset")
	BuildGuiFontOptions(Window, 0, "Gui")
	
	; Gui Options
	
	CtrlColour := Window.ControlColour = "" ? "Default" : Window.ControlColour
	Gui Editor:Color, , %CtrlColour%
	if(GetPropertyValue("GuiOptions", "Theme", Window) = "No")
		Gui Editor:-Theme
	
	for ControlType, Data in CurrentGui.UnqiueInstance
		if(!UniqueControls[Data.Hwnd])
			Data.Hwnd := 0
	
	; Recreate the controls and update handles (Hwnd)
	for ControlId, ControlHwnd in CurrentGui.ControlIndexes{
		ControlData := CurrentGui.Controls[ControlHwnd]
		CurrentGui.Controls.Remove(ControlHwnd, "")
		ControlHwnd := EditorWinCreateControlFromData(ControlData)
		
		if(ControlTypeOptions[ControlData.ControlType].IsUnique){
			Data := CurrentGui.UnqiueInstance[ControlData.ControlType]
			Data.Hwnd := ControlHwnd
			Data.Visible := 1
		}
	}
	
	CurrentControlData := CurrentGui.Window
	PropertyWinUpdateGroups(1)
	Gui Editor:Font
}

FileSave(SaveAs := 0){
	if(SaveAs or CurrentGui.Path = ""){
		Path := CurrentGui.Path = "" ? GetPropertyValue("Gui", "Title", CurrentGui.Window) ".gui" : CurrentGui.Path
		SetModalMode()
		Gui Main:+OwnDialogs
		FileSelectFile FilePath, S19, %Path%, Save File, Gui File (*.gui)
		SetModalMode(0)
		
		if(ErrorLevel)
			return
		if(SubStr(FilePath, -3) != ".gui")
			FilePath .= ".gui"
	}
	else
		FilePath := CurrentGui.Path
	
	File := FileOpen(FilePath, "w", "UTF-8")
	if(!IsObject(File)){
		ErrorMessage("There was an error opening the file")
		return
	}
	
	CurrentGui.Path := FilePath
	CurrentGui.Modified := 0
	FileWriteObject(File, CurrentGui)
	File.Close()
}

GetConstant(Name){
	if(ExportOptionNamedConstants){
		ExportOptionNamedConstantsList[Name] := 1
		return "%" Name "%"
	}
	else
		return Win32Constants[Name]
}

GetMenuItem(ItemPath){
	; Returns the actual text identifying a menu item used by AHK based on a simpler menu path (eg. File>Save)
	
	return % MainMenuAccKeys[ItemPath]
}

GetPropertyValue(PropertyGroup, PropertyName, Data, GetTextValue := 0){
	Property := PropertiesDef[PropertyGroup][PropertyName]
	if(GetTextValue = 2)
		PropertyClass.ControlType := Data.ControlType
	if(GetTextValue = 1 and Property.Get("Disable"))
		return "- NA -"
	
	Type := Property.Get("Type")
	Default := Property.Get("Default")
	Value := Data.HasKey(PropertyName) ? Data[PropertyName] : (IsObject(Default) ? (Data[PropertyName] := Default.Clone()) : Default)
	
	if(Type = "Control"){
		if(GetTextValue = 1){
			Page := Value >> 16
			Value := Value != "" and Value != 0
				? CurrentGui.Controls[CurrentGui.ControlIndexes[Value & 0xFFFF]].Id (Property.Get("AllowParentPage") and Page != 0 ? " > " Page : "")
				: "None"
		}
	}
	
	else if(Type = "Date"){
		Value := SubStr(Value, 2)
		if(Property.Get("NoTime") and Value != "")
			FormatTime Value, %Value%, yyyyMMdd
		if(GetTextValue = 1 and Value != "")
			FormatTime Value, %Value%, yyyy/MM/dd
	}
	
	else if(Type = "DateRange"){
		if(GetTextValue != "Raw"){
			DateMin := SubStr(Value[1], 2)
			DateMax := SubStr(Value[2], 2)
			
			if(Property.Get("NoTime")){
				if(DateMin != "")
					FormatTime DateMin, %DateMin%, yyyyMMdd
				if(DateMax != "")
					FormatTime DateMax, %DateMax%, yyyyMMdd
			}
			
			if(GetTextValue = 1){
				if(DateMin != "")
					FormatTime DateMin, %DateMin%, yyyy/MM/dd
				if(DateMax != "")
					FormatTime DateMax, %DateMax%, yyyy/MM/dd
			}
			
			Seperator := GetTextValue = 1 ? " - " : "-"
			IsDateRange := DateMax != "" and DateMin != ""
			NotEmpty := DateMax != "" or DateMin != ""
			Value := DateMin (IsDateRange or NotEmpty and !Property.Get("AllowSingle") ? Seperator : "") DateMax
		}
	}
	
	else if(Type = "Font"){
		if(GetTextValue = 1)
			Value := (Value.Size ? "(" Value.Size ") " : "") (Value.Name != "" ? Value.Name : "[Default]")
	}
	
	else if(Type = "IntToggle"){
		if(GetTextValue = 1)
			Value := (Property.Get("Default") = "" and Value[1] = "")
				? "[Default]"
				: Value[1] = "Yes" ? (Value[2] != "" ? Value[2] : "Yes") : "No"
	}
	
	else if(Type = "List"){
		if(Value = "")
			Value := Default
		if(GetTextValue = "Index")
			Value := Property.Get("ListIndexes")[Value]
	}
	
	else if(Type = "Text" or Type = "MultiText"){
		if(GetTextValue = 1){
			Value := RegExReplace(SubStr(Value, 1, 100), "[`n`r`t ]+", " ")
			if(Property.Get("GuiDelimiter"))
				StringReplace Value, Value, %GUI_DELIMITER%, % CurrentGui.Delimiter, 1
		}
	}
	
	else if(Type = "TextToggle"){
		if(GetTextValue = 1){
			Value := (Property.Get("Default") = "" and Value[1] = "")
				? "[Default]"
				: Value[1] = "Yes" ? (Value[2] != "" ? Value[2] : "Yes") : "No"
			
			if(Property.Get("GuiDelimiter"))
				StringReplace Value, Value, %GUI_DELIMITER%, % CurrentGui.Delimiter, 1
		}
	}
	
	else  if(Type = "Toggle"){
		Value := (Property.Get("Default") = "" and Value = "")
			? GetTextValue = 1 ? "[Default]" : ""
			: Value = "Yes" ? "Yes" : "No"
	}
	
	else if(Type = "XY"){
		if(GetTextValue = 1){
			SplitStr := (SplitString := Property.Get("SplitString")) ? SplitString : " x "
			X := Value[1], Y := Value[2]
			Value := X != "" or Y != ""
				? X SplitStr Y
				: ""
		}
	}
	
	return Value
}

GetTabNumber(ControlData, ByRef TabCount := 0){
	ControlData := !IsObject(ControlData) ? CurrentGui.Controls[ControlData] : ControlData
	TabNumber := GetPropertyValue("Tab2", "Choose", ControlData, 2)
	TabCount := GetListCount(ControlData.Text, GUI_DELIMITER)
	return TabCount = 0 ? 0 : (TabNumber > TabCount or TabNumber = "" or TabNumber < 1 ? 1 : TabNumber)
}

InitializePropertyDefinitions(){
	GroupControlTypes := {}
	
	; Recreate the property definitions as associative arrays so that a proprty can be accessed by its name and the properties can be displayed in the
	; same order that they were created
	for GroupName, PropertiesDefOptions in PropertiesDef{
		if(GroupName = "PropertyTypeOptions")
			continue
		
		GroupIndexes := []
		GroupIndex := 1
		PropertiesDef[GroupName] := Properties := {GroupIndexes: GroupIndexes}
		PropertiesList := PropertiesDefOptions.Properties
		if(PropertiesList[1] != "")
			Properties.GroupHeading := PropertiesList[1]
		
		; Copy across additional options
		if(PropertiesDefOptions.HasKey("Options"))
			for Key, Value in PropertiesDefOptions.Options
				Properties[Key] := Value
		
		; Find which types of controls these settings belong to
		ControlTypes := StrSplit(PropertiesDefOptions.ControlType, "|")
		TypeCount := ControlTypes.MaxIndex()
		
		while(TypeCount > 0){
			ControlType := ControlTypes[TypeCount - 1]
			GroupDisplayIndex := ControlTypes[TypeCount]
			
			Loop Parse, ControlType, `,
			{
				if(!GroupControlTypes.HasKey(A_LoopField))
					GroupControlTypes[A_LoopField] := {}
				
				GroupControlTypes[A_LoopField][GroupDisplayIndex] := GroupName
			}
			
			TypeCount -= 2
		}
		
		; Re-add each property with the property name as the key so that later on properties can easily be looked up by their name
		; Also initialize certain property options
		Index := 2
		Size := PropertiesList.MaxIndex()
		PropertyStack := [], PropertyStackIndex := 0
		while(Index <= Size){
			Property := PropertiesList[Index]
			Property.base := PropertyClass
			PropertyName := Property.Prop
			Properties[PropertyName] := Property
			Property.Remove("Prop", "")
			
			ParentType := !Property.HasKey("Type") ? (Property.Type := "Toggle") : Property.Type
			PropertyStack[++PropertyStackIndex] := Property
			
			if(Property.HasKey("ControlTypes")){
				Property.ControlTypes.Remove("")
				for ControlType, Options in Property.ControlTypes
					PropertyStack[++PropertyStackIndex] := Options
			}
			
			while(PropertyStackIndex){
				Property := PropertyStack[PropertyStackIndex--]
				Type := Property.Type ? Property.Type : ParentType
					
				if(Type = "Int" or Type = "IntToggle"){
					if(Type = "Int"){
						if(!Property.HasKey("Default"))
							Property.Default := ""
					}
					else if(!Property.HasKey("Default"))
						Property.Default := ["No", ""]
					else if(!IsObject(Property.Default))
						Property.Default := [Property.Default, ""]
					
					if(!Property.HasKey("Range"))
						Property.Range := [MIN_INT, MAX_INT]
					else{
						if(Property.Range[1] = "")
							Property.Range[1] := MIN_INT
						if(Property.Range[2] = "")
							Property.Range[2] := MAX_INT
					}
				}
				
				else if(Type = "XY"){
					if(!Property.HasKey("Default"))
						Property.Default := ["", ""]
					
					if(!Property.HasKey("Range"))
						Property.Range := [MIN_INT, MAX_INT]
					else{
						if(Property.Range[1] = "")
							Property.Range[1] := MIN_INT
						if(Property.Range[2] = "")
							Property.Range[2] := MAX_INT
					}
				}
				
				else if(Type = "Toggle"){
					if(!Property.HasKey("Default"))
						Property.Default := "No"
				}
				
				else if(Type = "List"){
					Property.ListIndexes := ListIndexes := {}
					ListOptions := StrSplit(Property.Options, "|")
					Options := Property.Options
					Loop Parse, Options, |
						ListIndexes[A_LoopField] := A_Index
					if((OptionValuesText := Property.OptionValues)){
						Property.OptionValues := OptionValues := {}
						Loop Parse, OptionValuesText, |
							OptionValues[A_Index] := A_LoopField
					}
				}
				
				else if(Type = "Font"){
					if(!Property.HasKey("Default"))
						Property.Default := {}
				}
				
				else if(Type = "TextToggle"){
					if(!Property.HasKey("Default"))
						Property.Default := ["No", ""]
					else if(!IsObject(Property.Default))
						Property.Default := [Property.Default, ""]
				}
				
				else if(Type = "DateRange"){
					if(!Property.HasKey("Default"))
						Property.Default := ["", ""]
					else if(!IsObject(Property.Default))
						Property.Default := [Property.Default, ""]
				}
				
				else{
					if(!Property.HasKey("Default"))
						Property.Default := ""
				}
			}
			
			GroupIndexes[GroupIndex++] := PropertyName
			Index++
		}
	}
	
	PropertiesDef.GroupControlTypes := GroupControlTypes
}

ParentControlAddChild(Parent, ChildIndex, Add := 1){
	ParentControl := ParentControls[Parent]
	ParentPage := Parent >> 16 & 0xFFFF
	ParentData := CurrentGui.Controls[ CurrentGui.ControlIndexes[Parent & 0xFFFF] ]
	ChildControls := ParentData._HasChildrenOnPage
	
	if(Add){
		if(!ParentControl)
			ParentControl := ParentControls[Parent] := {Visible: 1, Children: {}}
		ParentControl.Children[ChildIndex] := 1
		if(!ParentControl.Visible)
			GuiControl Editor:Hide, % CurrentGui.ControlIndexes[ChildIndex]
		if(!ChildControls)
			ChildControls := ParentData._HasChildrenOnPage := {}
		
		if(!ChildControls.HasKey(ParentPage))
			ChildControls[ParentPage] := 1
		else
			ChildControls[ParentPage]++
	}
	else{
		ParentControl.Children.Remove(ChildIndex, "")
		if(!ParentControl.Visible)
			GuiControl Editor:Show, % CurrentGui.ControlIndexes[ChildIndex]
		
		if(ChildControls and --ChildControls[ParentPage] = 0){
			ChildControls.Remove(ParentPage, "")
			if(ChildControls.MaxIndex() = "")
				ParentData.Remove("_HasChildrenOnPage")
		}
	}
}

ParentControlClearChildren(Parent){
	ParentControl := ParentControls[Parent]
	ParentControls.Remove(Parent, "")
	
	for ChildIndex, Value in ParentControl.Children{
		ControlHwnd := CurrentGui.ControlIndexes[ChildIndex]
		ControlData := CurrentGui.Controls[ControlHwnd]
		if(!ParentControl.Visible)
			GuiControl Editor:Show, %ControlHwnd%
		ControlData.Parent := 0
	}
}

ParentControlSetVisible(Parent, Visible, UpdateControls := 1){
	if(!(ParentData := ParentControls[Parent]))
		ParentData := ParentControls[Parent] := {Visible: Visible, Children: {}}
	
	if(UpdateControls and ParentData.Visible != Visible){
		for ChildIndex, Value in ParentData.Children
			GuiControl % Visible ? "Show" : "Hide", % ControlHwnd := CurrentGui.ControlIndexes[ChildIndex]
		
		ParentData.Visible := Visible
	}
}

RecreateControl(ControlHwnd, InsertInto := 0){
	ControlData := CurrentGui.Controls[ControlHwnd]
	DllCall("DestroyWindow", Ptr,ControlHwnd)
	ControlNewHwnd := EditorWinCreateControlFromData(ControlData, 0)
	CurrentGui.ControlIndexes[ControlData.Index] := ControlNewHwnd
	if(!InsertInto){
		CurrentGui.Controls.Remove(ControlHwnd, "")
		CurrentGui.Controls[ControlNewHwnd] := ControlData
	}
	else
		InsertInto[ControlNewHwnd] := ControlData
	OnEditorWinRecreateControl(ControlHwnd, ControlNewHwnd)
}

ResetCurrentGuiData(){
	; Initializes or resets the CurrentGui object
	
	UnqiueInstance := CurrentGui.UnqiueInstance ? CurrentGui.UnqiueInstance : 0
	CurrentGui := {Window: {ControlType: "Gui"}
		, Controls: {}, ControlCount: 0, ControlIndexes: [], NextIndex: 1, UnqiueInstance: {}
		, Delimiter: "|", RadioGroups: {}
		, Modified: 0, Path: ""}
	CurrentControlData := CurrentGui.Window
	
	CurrentGui.DefaultColour := GetPropertyValue("Gui", "C", CurrentGui.Window)
	if(UnqiueInstance)
		CurrentGui.UnqiueInstance := UnqiueInstance
}

SetApplicationHotKey(HotkeyName, Label := "", Options := "", WindowGroup := ""){
	Hotkey IfWinActive, % "ahk_group " (WindowGroup = "" ? "AppWindows" : WindowGroup)
	Hotkey %HotkeyName%, %Label%, %Options%
	Hotkey IfWinActive
}

SetGridSnap(Snap := true){
	Settings.SnapToGrid := Snap
	Menu ViewMenu, % Snap ? "Check" : "Uncheck", % GetMenuItem("View>SnapToGrid")
	ShowToolTip("Snap to Grid " (Snap ? "On" : "Off"), "Window", EditorWinHwnd)
	Gui Main:Default
	MainWinSetStatusBarIcon(Snap ? "grid-snap-on" : "grid-snap-off")
}

SetModalMode(Modal := 1){
	; Enables or disables all the application windows to create a "modal" dialog
	
	static EditorWinResizeHandleVisible := 0
		 , LastActive := 0
	
	if(Modal)
		LastActive := WinExist("A")
	else if(LastActive)
		WinActivate ahk_id %LastActive%
	
	Command := Modal ? "Disable" : "Enable"
	WinSet %Command%, , ahk_id %MainWinHwnd%
	WinSet %Command%, , ahk_id %EditorWinHwnd%
	WinSet %Command%, , ahk_id %PropertyWinHwnd%
}

SettingsLoad(){
	Settings := {}
	
	Loop Read, settings.ini
	{
		if( (Line := LTrim(A_LoopReadLine)) = "" )
			continue
		if(SubStr(Line, 1, 1) = ";"){
			;~ out("   // " SubStr(Line, 2))
			continue
		}
		;~ out("  '" Line "'")
		Index := InStr(A_LoopReadLine, "=")
		VarName := Trim(Index ? SubStr(A_LoopReadLine, 1, Index - 1) : A_LoopReadLine)
		VarValue := Index ? SubStr(A_LoopReadLine, Index + 1) : ""
		if(VarName != "")
			Settings[VarName] := VarValue
	}
	
	; Validate/Initialize settings
	if(Settings.GridSize < 2)
		Settings.GridSize := 2
	Settings.GridColour := ValidateHexColourValue(Settings.GridColour)
	;~ out("---------------")
	;~ out(Settings)
}

SettingsSave(){
	if(!IsObject(Settings)){
		ErrorMessage("There was an error saving the application settings")
		return
	}
	
	FileCopy %SETTINGS_FILE%.bak, %SETTINGS_FILE%.bak1, 1
	FileCopy %SETTINGS_FILE%, %SETTINGS_FILE%.bak, 1
	
	File := FileOpen(SETTINGS_FILE, "w")
	if(!IsObject(File)){
		ErrorMessage("There was an error saving the application settings")
		return
	}
	
	; Prepare settings
	Settings.GridColour := FormatHex(Settings.GridColour)
	
	File.Write("; Settings`n")
	
	For VarName, VarValue in Settings{
		;~ out(VarName "=" VarValue)
		File.Write(VarName "=" VarValue "`n")
	}
	
	File.Close()
}

ShowGrid(Show = true){
	Settings.ShowGrid := Show
	Menu ViewMenu, % Show ? "Check" : "Uncheck", % GetMenuItem("View>ShowGrid")
	WinSet Redraw, , ahk_id %EditorWinHwnd%
}

SnapToGrid(ByRef X, ByRef Y){
	X := Round(X / Settings.GridSize) * Settings.GridSize
	Y := Round(Y / Settings.GridSize) * Settings.GridSize
}

StoreControlStates(Controls, Mode := "store"){
	if( (Store := Mode = "store") )
		EditorControlState := {}, EditorControlStateGroup := []
	
	if(IsObject(Controls)){
		for ControlHwnd in Controls
			StoreControlState(ControlHwnd, Store)
	}
	else{
		StoreControlState(Controls, Store)
	}
}

StoreControlState(ControlHwnd, Store := 1){
	if(Store){
		if(EditorControlState.HasKey(ControlHwnd))
			return
		
		GuiControlGet Control, Editor:Pos, %ControlHwnd%
		EditorControlState[ControlHwnd] := {x: ControlX, y: ControlY, w: ControlW, h: ControlH}
		
		; Child controls
		ControlData := CurrentGui.Controls[ControlHwnd]
		if(ControlData._LinkChildren = 2){
			GuiControlGet Control, Editor:Pos, %ControlHwnd%
			X1 := ControlX, Y1 := ControlY, X2 := X1 + ControlW, Y2 := Y1 + ControlH
			for ChildControlHwnd, ControlData in CurrentGui.Controls{
				if(ChildControlHwnd = ControlHwnd)
					continue
				GuiControlGet Control, Editor:Pos, %ChildControlHwnd%
				ControlW += ControlX, ControlH += ControlY
				if(ControlX >= X1 and ControlW <= X2 and ControlY >= Y1 and ControlH <= Y2)
					StoreControlState(ChildControlHwnd)
					, EditorControlStateGroup.Insert(ChildControlHwnd)
			}
		}
		else if(ControlData._LinkChildren){
			for PageNumber in ControlData._HasChildrenOnPage
				for ChildIndex in ParentControls[ControlData.Index | (PageNumber << 16)].Children
					StoreControlState(CurrentGui.ControlIndexes[ChildIndex])
		}
	}
	else{
		State := EditorControlState[ControlHwnd]
		X := State.x, Y := State.y
		W := State.w, H := State.h
		GuiControl Editor:MoveDraw, %ControlHwnd%, x%X% y%Y% w%W% h%H%
		EditorControlState.Remove(ControlHwnd, "")
		EditorWinUpdateHighlight(ControlHwnd)
		ForceControlRedraw(ControlHwnd)
		
		ControlData := CurrentGui.Controls[ControlHwnd]
		ControlData.Position[1] := X, ControlData.Position[2] := Y
		ControlData.Size[1] := W, ControlData.Size[2] := H
		
		if(ControlHwnd = Selection.Current){
			EditorWinUpdateResizeHandles()
			OnPropertyWinPropertyChange("Common", "Position", "Common", "Size")
		}
		
		; Child controls
		ControlData := CurrentGui.Controls[ControlHwnd]
		if(ControlData._LinkChildren = 2){
			for Index, ControlHwnd in EditorControlStateGroup
				StoreControlState(ControlHwnd, 0)
		}
		else if(ControlData._LinkChildren){
			for PageNumber in ControlData._HasChildrenOnPage
				for ChildIndex in ParentControls[ControlData.Index | (PageNumber << 16)].Children
					StoreControlState(CurrentGui.ControlIndexes[ChildIndex], 0)
		}
	}
}

UpdateControlSize(ControlHwnd, UpdatePosition := 0){
	ControlData := CurrentGui.Controls[ControlHwnd]
	GuiControlGet Control, Editor:Pos, %ControlHwnd%
	ControlData.Size[1] := ControlW
	ControlData.Size[2] := ControlH
	
	if(UpdatePosition){
		ControlData.Position[1] := ControlX
		ControlData.Position[2] := ControlY
	}
}

; >> Settings window <<

SettingsWinCreateGui(){
	;~ Gui Settings:New, +HwndSettingsWinHwnd +OwnerMain +LabelOnSettingsWin, Settings - Gui Editor
	
	; TEMP:
	;~ Gui Settings:Show, w400 h500
}

; -- Events
; -----------------------------------------------------------------

OnAppExit:
{
	CloseApplication()
} return

OnMouseButtonDown(wParam, lParam, msg, hwnd){
	ReturnValue := ""
	
	OnEditorWinMouseDown(wParam, lParam, msg, hwnd, ReturnValue)
	
	if(ReturnValue = ""){
		if( (MouseDownFunc := CustomMouseDownEvents[hwnd]) ){
			if(MouseDownFunc.Threaded){
				if(MouseDownFunc.Init)
					MouseDownFunc.Threaded.(MouseDownFunc.Param)
				CallThreadedFunction(MouseDownFunc.Threaded, MouseDownFunc.Param)
			}
			else
				MouseDownFunc.()
		}
	}
	
	return ReturnValue
}

OnMouseWheel:
{
	MouseGetPos, , , WinHwnd, , 2
	if(WinHwnd = PropertyWinHwnd)
		OnPropertyWinMouseWheel(A_ThisHotkey)
	
} return