; -- Initialize variables
; -----------------------------------------------------------------
{
	; Restricted input types
	global RI_INTEGER			:= 0
		, RI_INTEGER_POSITIVE	:= 1
		, RI_LABEL				:= 2
		, RI_VARIABLE			:= 3
		; Restricted input regular expressions
		, RI_INTEGER_REGEX				:= "[-0-9]"
		, RI_INTEGER_REGEX_I			:= "[^-0-9]"
		, RI_INTEGER_POSITIVE_REGEX		:= "[0-9]"
		, RI_INTEGER_POSITIVE_REGEX_I	:= "[^0-9]"
		, RI_LABEL_REGEX				:= "[^`t`n`r ``,]"
		, RI_LABEL_REGEX_I				:= "[`t`n`r ``,]"
		, RI_VARIABLE_REGEX				:= "[0-9a-zA-Z#_@$]"
		, RI_VARIABLE_REGEX_I			:= "[^0-9a-zA-Z#_@$]"
	
	global ControlTooltips := {}	; Maps tooltip text to a control's Hwnd
		 , RestrictedInputs := {}
		 , ThreadedFunc := {}
}

class _Array{
	IsArray(){
		return 1
	}
}

Array(prm*){
	prm.base := _Array
	return prm
}

; -- Functions
; -----------------------------------------------------------------

ArrayJoin(ArrayVar, ByRef Result, Delimiter := "|"){
	if((Count := ArrayVar.MaxIndex()) = 0){
		Result := ""
		return
	}
	
	Result := ArrayVar[1]
	Index := 2
	
	while(Index <= Count)
		Result .= Delimiter ArrayVar[Index++]
}

BinarySearch(ArrayVar, Value){
	Size := ArrayVar.MaxIndex()
	Low := 1, High := (Size != "" ? Size : 0)
	While(Low <= High){
		Mid := Floor((Low + High) / 2) ; round to lower integer
		if(ArrayVar[Mid] > value)
			High := Mid - 1
		else if(ArrayVar[Mid] < value)
			Low := Mid + 1
		else
			return Mid
	}
	
	return 0
}

BinarySearchLessThan(ArrayVar, Value){
	Size := ArrayVar.MaxIndex()
	Low := 1, High := (Size != "" ? Size : 0)
	While(Low <= High){
		Mid := Floor((Low + High) / 2) ; round to lower integer
		if(ArrayVar[Mid] > value)
			High := Mid - 1
		else if(ArrayVar[Mid] < value)
			Low := Mid + 1
		else
			return Mid + 1
	}
	
	return Low
}

CallThreadedFunction(Function, Param := "", t := 0){
	ThreadedFunc.Function := Function
	ThreadedFunc.Param := Param
	SetTimer OnThreadedFunctionCall, -%t%
}

ClientToScreen(Hwnd, ByRef X, ByRef Y){
	VarSetCapacity(Rect, 8, 0)
	NumPut(X, Rect, 0, "Int")
	NumPut(Y, Rect, 4, "Int")
	DllCall("ClientToScreen", Ptr, Hwnd, Ptr, &Rect)
	X := NumGet(Rect, 0, "Int")
	Y := NumGet(Rect, 4, "Int")
}

ColourNameToRGB(Colour, RGBToName := 0){
	static NameToColourMap := {Black: 0x000000, Green: 0x008000, Silver: 0xC0C0C0, Lime: 0x00FF00, Gray: 0x808080, Olive: 0x808000, White: 0xFFFFFF, Yellow: 0xFFFF00
							 , Maroon: 0x800000, Navy: 0x000080, Red: 0xFF0000, Blue: 0x0000FF, Purple: 0x800080, Teal: 0x008080, Fuchsia: 0xFF00FF, Aqua: 0x00FFFF}
		 , ColourToNameMap := {0x000000: "Black", 0x008000: "Green", 0xC0C0C0: "Silver", 0x00FF00: "Lime", 0x808080: "Gray", 0x808000: "Olive", 0xFFFFFF: "White", 0xFFFF00: "Yellow"
							 , 0x800000: "Maroon", 0x000080: "Navy", 0xFF0000: "Red", 0x0000FF: "Blue", 0x800080: "Purple", 0x008080: "Teal", 0xFF00FF: "Fuchsia", 0x00FFFF: "Aqua"}
	
	if(RGBToName)
		return ColourToNameMap[Colour]
	
	return NameToColourMap[Colour]
}

ColourValueToRGBHex(Colour){
	Colour := SubStr(FormatHex(Colour), 3)
	Length := StrLen(Colour)
	while(Length++ < 6)
		Colour := "0" Colour
	return "0x" Colour
}

DeepCloneObject(Object, Objs := 0){
	if(!Objs)
		Objs := {}
	
	Obj := Object.Clone()
	Objs[&Object] := Obj ; Save this new array
	
	for Key, Val in Obj
		if(IsObject(Val)) ; If it is a subarray
			Obj[Key] := Objs[&Val] ; If we already know of a refrence to this array
				? Objs[&Val] ; Then point it to the new array
				: DeepCloneObject(Val,Objs) ; Otherwise, clone this sub-array
	
	return Obj
}

ExpandEnvironmentStrings(Path){
	VarSetCapacity(PathFull, 2000, 0)
	DllCall("ExpandEnvironmentStrings", Str,Path, Str,PathFull, Int,1999, "Cdecl Int")
	return PathFull
}

FileReadArray(File){
	Count := File.ReadUShort()
	
	Data := []
	while(Count > 0){
		Index := File.ReadUShort()
		Data[Index] := File.Read( File.ReadUShort() )
		Count--
	}
	
	return Data
}

FileReadObject(File){
	Data := {}
	
	Loop
	{
		KeyLength := File.ReadUChar()
		if(!KeyLength)
			break
		
		Key := File.Read(KeyLength - 1)
		ValueType := File.ReadUChar()
		
		if(ValueType = 10)
			ValueType := File.ReadUChar(), ReadHexValue := 1
		else
			ReadHexValue := 0
		
		if(ValueType = 0)
			Value := FileReadArray(File)
		else if(ValueType = 1)
			Value := FileReadObject(File)
		else if(ValueType = 2)
			Value := File.ReadUChar()
		else if(ValueType = 3)
			Value := File.ReadUShort()
		else if(ValueType = 4)
			Value := File.ReadUInt()
		else if(ValueType = 5)
			Value := File.ReadChar()
		else if(ValueType = 6)
			Value := File.ReadShort()
		else if(ValueType = 7)
			Value := File.ReadInt()
		else if(ValueType = 8)
			Value := File.ReadDouble()
		else if(ValueType = 9)
			Value := File.Read( File.ReadUShort() )
		
		Data[Key] := ReadHexValue ? FormatHex(Value) : Value
	}
	
	return Data
}

FileWriteArray(File, Data, Count := -1){
	if(Count = -1){
		Count := 0
		for Index in Data
			Count++
	}
	
	File.WriteUShort(Count)
	
	for Index, Value in Data{
		File.WriteUShort(Index)
		File.WriteUShort(StrLen(Value))
		File.Write(Value)
		if(--Count = 0)
			break
	}
}

FileWriteObject(File, Data, Indent := ""){
	for Key, Value in Data{
		File.WriteUChar(StrLen(Key) + 1)
		File.Write(Key)
		
		if(Value.IsArray()){
			File.WriteUChar(0)
			FileWriteArray(File, Value)
		}
		else if(IsObject(Value)){
			File.WriteUChar(1)
			FileWriteObject(File, Value, Indent "  ")
		}
		else{
			if Value is Integer
			{
				if(SubStr(Value, 1, 2) = "0x")
					File.WriteUChar(10)
				if(Value >= 0){
					if(Value <= 0xFF)
						File.WriteUChar(2), File.WriteUChar(Value)
					else if(Value <= 0xFFFF)
						File.WriteUChar(3), File.WriteUShort(Value)
					else
						File.WriteUChar(4), File.WriteUInt(Value)
				}
				else{
					if(Value >= -128)
						File.WriteUChar(5), File.WriteChar(Value)
					else if(Value >= -32768)
						File.WriteUChar(6), File.WriteShort(Value)
					else
						File.WriteUChar(7), File.WriteInt(Value)
				}
			}
			else if Value is Float
				File.WriteUChar(8), File.WriteDouble(Value)
			else
				File.WriteUChar(9), File.WriteUShort(StrLen(Value)), File.Write(Value)
		}
	}
	File.WriteUChar(0)
}

FindControlParentWindow(Hwnd, ByRef ParentHwnd){
	ParentHwnd := DllCall("GetAncestor", Ptr,Hwnd, UInt,GA_ROOT)
}

ForceControlRedraw(ControlHwnd, Full := 0){
	if(Full){
		GuiControlGet ControlPos, Pos, %ControlHwnd%
		GuiControl Move, %ControlHwnd%, % "w" (ControlPosW + 1)
		GuiControl Move, %ControlHwnd%, % "w" ControlPosW
	}
	else
		WinSet Redraw, , ahk_id %ControlHwnd%
}

FormatHex(InNumber){
	OldFormat := A_FormatInteger
	SetFormat, Integer, Hex
	FormattedNumber := InNumber + 0
	SetFormat, Integer, %OldFormat%
	return FormattedNumber
}

GetClientSize(Hwnd, ByRef Width, ByRef Height){
	VarSetCapacity(Rect, 16)
	DllCall("GetClientRect", Ptr,Hwnd, Ptr,&Rect)
	Width  := NumGet(Rect, 8, "Int")
	Height := NumGet(Rect, 12, "Int")
}

GetListCount(ByRef ListString, Delimiter){
	Index := 1, DelimiterCount := 0
	
	while(Index := InStr(ListString, Delimiter, 1, Index))
		DelimiterCount++
		, Index++
	
	return ListString != "" ? DelimiterCount + 1 : DelimiterCount
}

GetScrollInfo(Hwnd, ScrollBar, ByRef ScrollInfo){
	VarSetCapacity(ScrollInfo, 28, 0)
	NumPut(28, ScrollInfo, 0, "UInt") ; cbSize
	NumPut(SIF_ALL, ScrollInfo, 4, "UInt") ; fMask
	return DllCall("GetScrollInfo", UInt,Hwnd, Int,ScrollBar, Ptr,&ScrollInfo)
}

GetTextSize(TextHwnd, NewText, ByRef NewWidth := 0, ByRef NewHeight := 0){
	DC := DllCall("GetDC", Ptr,TextHwnd)
	SendMessage WM_GETFONT, 0, 0, , ahk_id %TextHwnd%
	FontHwnd := ErrorLevel
	OriginalFont := DllCall("SelectObject", Ptr,DC, Ptr,FontHwnd, Ptr)
	
	VarSetCapacity(RECT, 16, 0)
	DllCall("DrawText", Ptr,DC, Str,NewText, Int,-1, Ptr,&RECT, UInt,DT_CALCRECT)
	NewWidth  := NumGet(RECT, 8, "Int")
	NewHeight := NumGet(RECT, 12, "Int")
	
	DllCall("SelectObject", Ptr,DC, Ptr,OriginalFont, Ptr)
	DllCall("DeleteObject", Ptr,FontHwnd)
	DllCall("ReleaseDC", Ptr,DC)
}

GetWindowRect(Hwnd, ByRef X, ByRef Y, ByRef Width, ByRef Height){
	VarSetCapacity(Rect, 16)
	DllCall("GetWindowRect", Ptr,Hwnd, Ptr,&Rect)
	X := NumGet(Rect, 0, "Int")
	Y := NumGet(Rect, 4, "Int")
	Width  := NumGet(Rect, 8, "Int")
	Height := NumGet(Rect, 12, "Int")
}

MapWindowPoints(hWndFrom, hWndTo, ByRef X, ByRef Y){
	VarSetCapacity(lpPoints, 8, 0)
	NumPut(X, lpPoints, 0, "Int")
	NumPut(Y, lpPoints, 4, "Int")
	Result := DllCall("MapWindowPoints", Ptr,hWndFrom, Ptr,hWndTo, Ptr,&lpPoints, UInt,1)
	X := NumGet(lpPoints, 0, "Int")
	Y := NumGet(lpPoints, 4, "Int")
	return Result
}

PickIconDlg(ByRef IconPath, ByRef IconIndex, ParentHwnd := 0){
	VarSetCapacity(wIconPath , 260 * 2)
	if(!A_IsUnicode)
		DllCall("MultiByteToWideChar", UInt,0, UInt,0, Str,IconPath, Int,-1, Str,wIconPath, Int,260)
	else
		wIconPath := IconPath
	
	if(!DllCall("shell32\PickIconDlg", Ptr,ParentHwnd, Str,wIconPath, UInt,260, IntP,IconIndex))
		return 0
	
	if(!A_IsUnicode){
		VarSetCapacity(sIconPath, 260, sIconPath)
		DllCall("WideCharToMultiByte", UInt,0, UInt,0, Str,wIconPath, Int,-1, Str,sIconPath, Int,1025, UInt,0, UInt,0)
	}
	else
		IconPath := wIconPath
	
	return 1
}

PrintObjFormatted(Obj, TransformFunc := 0, Transforms := 0, Depth=12, CurIndent="", Path := ""){
	if(!IsObject(Obj)){
		if Obj is number
			return Obj
		if(RegExMatch(Obj, "i)Hwnd\[0x[0-9A-Fa-f]+\]"))
			return Obj
		
		return """" RegExReplace(Obj, "(`n|`r`n)", """$1" CurIndent "    """) """"
	}
	
	ToReturn := ""
	if(Depth > 1)
	for Key, Value in Obj{
		ItemPath := Path "/" Key
		
		for Index, Filter in Transforms{
			if(RegExMatch(ItemPath, Filter)){
				TransformFunc.(Key, Value, Index)
				break
			}
		}
		
		ToReturn .= CurIndent . Key . " = " PrintObjFormatted(Value, TransformFunc, Transforms, Depth-1, CurIndent . "|  ", ItemPath) . "`n"
	}
	
	if(Depth = 1)
		ToReturn := (Depth < 12 ? "`n" : "") CurIndent "..."
	else if(ToReturn = "")
		ToReturn := "[EMPTY]"
	else 
		ToReturn := (Depth < 12 ? "`n" : "") ToReturn
	
	return RTrim(ToReturn, "`n") 
}

RestrictInput(EditHwnd, Type){
	RestrictedInputs[EditHwnd] := Type
	SubClassWindow(EditHwnd, "OnRestrictedEditWindowProc")
}

ScreenToClient(ByRef X, ByRef Y, Hwnd){
	VarSetCapacity(Rect, 8, 0)
	NumPut(X, Rect, 0, "Int")
	NumPut(Y, Rect, 4, "Int")
	DllCall("ScreenToClient", Ptr, Hwnd, Ptr, &Rect)
	X := NumGet(Rect, 0, "Int")
	Y := NumGet(Rect, 4, "Int")
}

ScrollWindow(Hwnd, Msg, wParam, lParam, Bottom = 0){
	ScrollStep := PropertyWinRowHeight ; Scroll an entire row at a time
	ScrollBar := Msg = WM_VSCROLL
	
	if(!GetScrollInfo(Hwnd, ScrollBar, ScrollInfo))
		return
	
	nPage  := NumGet(ScrollInfo, 16, "UInt")
	NewPos := OldPos := NumGet(ScrollInfo, 20, "Int") ; nPos
	
	if(wParam = "ScrollTo"){
		Top := lParam
		if(Bottom > nPage)
			NewPos += Bottom - nPage
		else if(Top < 0)
			NewPos += Top
		else
			return
	}
	else{
		Action := wParam & 0xFFFF
		if(Action = SB_LINEUP)
			NewPos -= ScrollStep
		else if(Action = SB_LINEDOWN)
			NewPos += ScrollStep
		else if(Action = SB_PAGEUP)
			NewPos -= nPage
		else if(Action = SB_PAGEDOWN)
			NewPos += nPage
		else if(Action = SB_THUMBTRACK || Action = SB_THUMBPOSITION)
			NewPos := wParam>>16
		else if(Action = SB_TOP)
			NewPos := NumGet(ScrollInfo, 8, "Int") ; nMin
		else if(Action = SB_BOTTOM)
			NewPos := NumGet(ScrollInfo, 12, "Int") ; nMax
		else
			return
	}
	
	Min := NumGet(ScrollInfo, 8, "Int") ; nMin
	Max := NumGet(ScrollInfo, 12, "Int") - nPage ; nMax - nPage
	NewPos := NewPos > Max ? Max : NewPos
	NewPos := NewPos < Min ? Min : NewPos
	
	X := Y := 0
	if(ScrollBar = 0) ; SB_HORZ
		X := OldPos - NewPos
	else
		Y := OldPos - NewPos
	; Scroll contents of window and invalidate uncovered area.
	DllCall("ScrollWindowEx", Ptr,Hwnd, Int,x, Int,y, Ptr,0, Ptr,0, Ptr,0, Ptr,0, UInt,SW_REDRAW)
	; Update scroll bar.
	NumPut(NewPos, ScrollInfo, 20, "Int") ; nPos
	DllCall("SetScrollInfo", Ptr,Hwnd, Int,ScrollBar, Ptr,&ScrollInfo, Int,1)
}

SetBitFlag(ByRef Value, Flag, On := 1){
	if(On)
		Value |= Flag
	else
		Value &= ~Flag
}

SetControlTooltip(Hwnd, Text, Pos := 0){
	; Adds a simple tooltip to any control
	
	WinGetClass ClassNN, ahk_id %Hwnd%
	if(ClassNN = "Static")			; Static control doesn't report WM_MOUSEMOVE without this flag.
		WinSet Style, +%SS_NOTIFY%, ahk_id %Hwnd%
	else if(ClassNN = "ComboBox")	; While in combo, it happens in edit owned by the combo
		ControlGet Hwnd, Hwnd, , Edit1, ahk_id %Hwnd%
	
	ControlTooltips[Hwnd] := text
	if(Pos)
		ControlTooltips[(Hwnd + 0) "Pos"] := 1
	SubClassWindow(Hwnd, "OnControlTooltipWindowProc")
	
	return
}

ShowControlInfoToolTip(ControlHwnd, X := "", Y := "", Time := 0){
	; Displays information about the given control in a tooltip (x, y, width and height)
	; If X = "auto" the tooltip will automatically be displayed below the control
	
	GetWindowRect(ControlHwnd, ControlX, ControlY, ControlW, ControlH)
	ControlW -= ControlX, ControlH -= ControlY
	ParentHwnd := DllCall("GetAncestor", Ptr,ControlHwnd, UInt,GA_ROOT)
	ScreenToClient(ControlX, ControlY, ParentHwnd)
	
	if(X = "auto"){
		X := ControlX, Y := ControlY + ControlH + 2
		ClientToScreen(ParentHwnd, X, Y)
	}
	
	ShowToolTip("x: " ControlX "`ny: " ControlY "`nWidth: " ControlW "`nHeight: " ControlH, X, Y, Time)
}

ShowToolTip(Text, X := "", Y := "", Time := 1000){
	; Displays a tooltip for a certain amount of time
	
	static CurrentToolTip := ""
	
	if(X = "" or Y = ""){
		MouseGetPos MouseX, MouseY
		X := X = "" ? MouseX + 2 : X
		Y := Y = "" ? MouseY + 20 : Y
	}
	
	; If X = "Ed", display the tooltip at the top-left of the editor window client area
	if(X = "Window"){
		WinHwnd := Y, X := Y := 0
		ClientToScreen(WinHwnd, X, Y)
		X += 4, Y += 4
	}
	
	; Only redisplay if the the text or position of the tooltip has changed
	NewToolTip := Text "|" X "|" Y
	if(NewToolTip != CurrentToolTip){
		ToolTip %Text%, %X%, %Y%
		CurrentToolTip := NewToolTip
	}
	
	if(Time > 0)
		SetTimer OnToolTipComplete, -%Time%
	else
		SetTimer OnToolTipComplete, Off
}

SubClassWindow(Hwnd, FunctionName){
	GetWindowLong := A_PtrSize = 8 ? "GetWindowLongPtr" : "GetWindowLong"
	SetWindowLong := A_PtrSize = 8 ? "SetWindowLongPtr" : "SetWindowLong"
	
	if FunctionName is not integer
		WindowProcOld := DllCall(GetWindowLong, Ptr, Hwnd, Int, GWL_WNDPROC, Ptr)
		, WindowProcNew := RegisterCallback(FunctionName, "", 4, WindowProcOld)
	else
		WindowProcNew := FunctionName
		, WindowProcOld := 0
	
	DllCall(SetWindowLong, Ptr,Hwnd, Int,GWL_WNDPROC, Ptr,WindowProcNew, Ptr)
	
	return WindowProcOld
}

UpdateScrollBars(WindowHwnd, WindowWidth, WindowHeight){
	; Calculate scrolling area
	
	Gui, %WindowHwnd%:Default
	Left := Top := 0
	Right := Bottom := 0
	ControlHwnd := 0
	
	Loop{
		if( !(ControlHwnd := DllCall("FindWindowEx", Ptr,WindowHwnd, Ptr,ControlHwnd, Ptr,0, Ptr,0)) )
			break
		
		ControlGet ControlVisible, Visible, , , ahk_id %ControlHwnd%
		if(!ControlVisible)
			continue
		GuiControlGet Control, Pos, %ControlHwnd%
		if(ErrorLevel)
			continue
		
		if(ControlX < Left)
			Left := ControlX
		if(ControlY < Top)
			Top := ControlY
		if(ControlX + ControlW > Right)
			Right := ControlX + ControlW
		if(ControlY + ControlH > Bottom)
			Bottom := ControlY + ControlH
	}
	
	ScrollWidth  := Right - Left
	ScrollHeight := Bottom - Top
	
	; Initialize SCROLLINFO
	VarSetCapacity(ScrollInfo, 28, 0)
	NumPut(28, ScrollInfo, 0, "UInt") ; cbSize
	NumPut(SIF_RANGE | SIF_PAGE, ScrollInfo, 4, "Uint") ; fMask
	
	; Update horizontal scroll bar.
	NumPut(ScrollWidth, ScrollInfo, 12, "Int") ; nMax
	NumPut(WindowWidth, ScrollInfo, 16, "UInt") ; nPage
	DllCall("SetScrollInfo", Ptr,WindowHwnd, UInt,SB_HORZ, Ptr,&ScrollInfo, Int,1)
	
	; Update vertical scroll bar.
    NumPut(SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL, ScrollInfo, 4, "UInt") ; fMask
	NumPut(ScrollHeight, ScrollInfo, 12, "Int") ; nMax
	NumPut(WindowHeight, ScrollInfo, 16, "UInt") ; nPage
	DllCall("SetScrollInfo", Ptr,WindowHwnd, UInt,SB_VERT, Ptr,&ScrollInfo, Int,1)
	
	; Make sure the window's scroll position is within the new scroll range
	X := 0, Y := 0
	if(Left < 0 and Right < WindowWidth)
		X := Abs(Left) > WindowWidth - Right ? WindowWidth - Right : Abs(Left)
	if(Top < 0 and Bottom < WindowHeight)
		Y := Abs(Top) > WindowHeight - Bottom ? WindowHeight - Bottom : Abs(Top)
	if(X || Y)
		DllCall("ScrollWindowEx", Ptr,WindowHwnd, Int,X, Int,Y, Ptr,0, Ptr,0, Ptr,0, Ptr,0, UInt,SW_REDRAW)
}

ValidateHexColourValue(Colour){
	OutColour := Colour
	
	; Expand a 3 (RGB) or 4 (ARGB) digit hex colour to a 6 (RRGGBB) and 8 (AARRGGBB) hex value, eg. FA0 > FFAA00
	; The colour may optionally start with 0x
	if(InStr(Colour, "0x") != 1){
		if(StrLen(Colour) = 3 or StrLen(Colour) = 4)
			OutColour := SubStr(Colour, 1, 1) SubStr(Colour, 1, 1) SubStr(Colour, 2, 1) SubStr(Colour, 2, 1) SubStr(Colour, 3, 1) SubStr(Colour, 3, 1)
		if(StrLen(Colour) = 4)
			OutColour .= SubStr(Colour, 4, 1) SubStr(Colour, 4, 1)
		OutColour := "0x" OutColour
	}
	else if(StrLen(Colour) = 5 or StrLen(Colour) = 6){
		OutColour := "0x" SubStr(Colour, 3, 1) SubStr(Colour, 3, 1) SubStr(Colour, 4, 1) SubStr(Colour, 4, 1) SubStr(Colour, 5, 1) SubStr(Colour, 5, 1)
		if(StrLen(Colour) = 6)
			OutColour .= SubStr(Colour, 6, 1) SubStr(Colour, 6, 1)
	}
	
	if(StrLen(OutColour) < 10){ ; If the colour does not have an alpha component set it to FF
		OutColour += 0 ; Force an integer
		OutColour |= 0xFF000000
	}
	
	return OutColour + 0
}

ValidateIntegerValue(ByRef Value, AllowEmpty := 1){
	StrValue := Trim(Value)
	StringReplace StrValue, StrValue, %A_Space%, , 1
	if(AllowEmpty and StrValue = "")
		Value := ""
	else{
		Value := StrValue
		Value += 0
	}
}

; -- Events
; -----------------------------------------------------------------

UtilitiesSubroutinesDef(){
	return
	
	OnThreadedFunctionCall:
		OnThreadedFunctionCall()
	return
	
	OnToolTipComplete:
		OnToolTipComplete()
	return
}

OnControlTooltipWindowProc(hwnd, msg, wParam, lParam){
	Critical
	static CurrentControl =
	
	if(msg = WM_MOUSEMOVE and CurrentControl != hwnd){
		Pos := ControlTooltips[(hwnd + 0) "Pos"]
		
		; If Pos = 1 then the tooltip is displayed directly underneath the control, otherwise it is displayed underneath the mouse.
		; The Pos option can be set when creating the control's tooltip with the SetControlTooltip function
		if(Pos){
			FindControlParentWindow(hwnd, ParentWindow)
			GetClientSize(hwnd, ControlW, ControlH)
			ControlX := ControlY := 0
			ClientToScreen(hwnd, ControlX, ControlY)
			ControlY += ControlH + 2
		}
		else{
			ControlX =
			ControlY =
		}
		
		Tooltip % ControlTooltips[hwnd], %ControlX%, %ControlY%, 2
		CurrentControl := hwnd
	}
	
	if(msg = WM_MOUSELEAVE){
		CurrentControl =
		Tooltip, , , , 2
	}
	
	return DllCall("CallWindowProc", Ptr,A_EventInfo, Ptr,hwnd, UInt,msg, Ptr,wParam, Ptr,lParam)
}

OnRestrictedEditWindowProc(hwnd, msg, wParam, lParam){
	Critical
	
	if(msg = WM_CHAR){
		Type := RestrictedInputs[hwnd]
		if(wParam = VK_PASTE){
			if(Type = RI_INTEGER)
				TypeRegEx := RI_INTEGER_REGEX_I
			if(Type = RI_INTEGER_POSITIVE)
				TypeRegEx := RI_INTEGER_POSITIVE_REGEX_I
			else if(Type = RI_LABEL)
				TypeRegEx := RI_LABEL_REGEX_I
			else if(Type = RI_VARIABLE)
				TypeRegEx := RI_VARIABLE_REGEX_I
			
			; Handle the PASTE manually becuase attempting to modify the clipboard here (through the built-in Clipboard command) will fail when the source of the data is
			; this or another edit with the OnRestrictedEditWindowProc custom WindowProc
			PasteString := RegExReplace(Clipboard, TypeRegEx, "")
			SendMessage EM_REPLACESEL, 1, &PasteString, , ahk_id %hwnd%
			return 0
		}
		else if(wParam != VK_BACKSPACE and wParam != VK_COPY and wParam != VK_CUT and wParam != VK_UNDO){
			if(Type = RI_INTEGER)
				TypeRegEx := RI_INTEGER_REGEX
			if(Type = RI_INTEGER_POSITIVE)
				TypeRegEx := RI_INTEGER_POSITIVE_REGEX
			else if(Type = RI_LABEL)
				TypeRegEx := RI_LABEL_REGEX
			else if(Type = RI_VARIABLE)
				TypeRegEx := RI_VARIABLE_REGEX
			
			if(!RegExMatch(Chr(wParam), TypeRegEx))
				return 0
		}
	}
	
	return DllCall("CallWindowProc", Ptr,A_EventInfo, Ptr,hwnd, UInt,msg, Ptr,wParam, Ptr,lParam)
}

OnThreadedFunctionCall(){
	ThreadedFunc.Function.(ThreadedFunc.Param)
	ThreadedFunc.Function := ThreadedFunc.Param := ""
}

OnToolTipComplete(){
	SetTimer OnToolTipComplete, Off
	ToolTip
}