﻿/*
 Function:  Font
			 (See Dlg_font.png)				

 Parameters:
            Name	- Initial font,  output.
            Size	- Initial Size, output.
            Style	- Initial style, output.
            Color	- Initial text color, output.
			Effects	- Set to false to disable effects (strikeout, underline, color).
            hGui	- Parent's handle, affects position.

  Returns:
            False if user canceled the dialog or if error occurred.
 */
Dlg_Font(ByRef Name, ByRef Size, ByRef Style, ByRef Colour, Effects := 1, hGui := 0){
	static LogFont
	
	PtrType := A_PtrSize ? "Ptr" : "UInt"
	, PtrSize := A_PtrSize ? A_PtrSize : 4
	
	LogPixels := DllCall("GetDeviceCaps", PtrType,DllCall("GetDC", PtrType,hGui), "UInt",90)	;LOGPIXELSY
	VarSetCapacity(LogFont, 60, 0)
	Effects := 0x041 + (Effects ? 0x100 : 0)  ;CF_EFFECTS = 0x100, CF_SCREENFONTS=1, CF_INITTOLOGFONTSTRUCT = 0x40
	
	; Convert from RGB to BGR - Include the alpha value (0xFF000000) even though it is not used
	Colour := (Colour & 0xFF000000) | ((Colour & 0xFF) << 16) + (Colour & 0xFF00) + ((Colour >> 16) & 0xFF)
	
	; Set initial name
	if(PtrSize = 8)
		StrPut(Name, &LogFont + 28)
	else
		StrPut(Name, &LogFont + 28, 32, "UTF-8")
	
	;set intial data
	Size := Size = "" ? 8 : Size
	NumPut(InStr(Style, "bold") ? 700 : 400, LogFont, 16, "UInt")
	if(InStr(Style, "italic"))
		NumPut(1, LogFont, 20, "Char")
	if(InStr(Style, "underline"))
		NumPut(1, LogFont, 21, "Char")
	if(InStr(Style, "strikeout"))
		NumPut(1, LogFont, 22, "Char")
	NumPut(Size * LogPixels / 72, LogFont, 0, "Int")
	
	; Build CHOOSEFONT structure
	StructSize := VarSetCapacity(ChooseFont, PtrSize = 8 ? 103 : 60, 0)
	, NumPut(StructSize,	ChooseFont, 00, "UInt")					; DWORD lStructSize
	, NumPut(hGui,			ChooseFont, PtrSize, PtrType)			; HWND hwndOwner (makes dialog "modal").
	, NumPut(&LOGFONT,		ChooseFont, 3 * PtrSize, PtrType)		; LPLOGFONT lpLogFont
	, NumPut(Effects,		ChooseFont, 04 + 4 * PtrSize, "UInt")	; DWORD Flags
	, NumPut(Colour,		ChooseFont, 08 + 4 * PtrSize, "UInt")	; rgbColors
	
	Result := DllCall("comdlg32\ChooseFont" . (A_IsUnicode ? "W" : "A"), PtrType, &ChooseFont, "cdecl")  ; Display the dialog. ; NOTE: omitting the encoding ternary makes it fail on AHK classic
	if(!Result)
		return 0
	
	; Font name
	if(PtrSize = 8)
		Name := StrGet(&LogFont + 28)
	else
		Name := StrGet(&LogFont + 28, 32, "UTF-8")
	Size := NumGet(ChooseFont, 4 * PtrSize, "Int") // 10
	
	; Color
	Colour := NumGet(ChooseFont, 08 + 4 * PtrSize, "UInt")
	; Convert from BGR to RGB
	Colour := (Colour & 0xFF000000) | (Colour & 0xFF00) + ((Colour & 0xFF0000) >> 16) + ((Colour & 0xFF) << 16) 
	
	; Styles
	Style := ""
	if(NumGet(LogFont, 16, "UInt") >= 700)
		Style .= "bold "
	if(NumGet(LogFont, 20, "UChar"))
		Style .= "italic "
	if(NumGet(LogFont, 21, "UChar"))
		Style .= "underline "
	if(NumGet(LogFont, 22, "UChar"))
		Style .= "strikeout "
	
	return 1
}